import os, sys
from argparse import ArgumentParser


ENV_VAR = 'PRICE_VAL_ENG_SETTINGS_MODULE'

def handle_default_options(options):
    """
    Include any default options to the all commands 
    """
    if options.settings:
        os.environ[ENV_VAR] = options.settings
    

class BaseCommand:
    
    def get_version(self):
        import price_val_engine
        return price_val_engine.get_version()

    def short_desc(self):
        return ""
    
    def long_desc(self):
        return self.short_desc()
    
    def help(self):
        return self.long_desc()
    
    def create_parser(self, prog_name, subcommand, **kwargs):
        """
        Create and return the ``ArgumentParser`` which will be used to
        parse the arguments to this command.
        """
        parser = ArgumentParser(
            prog='%s %s' % (os.path.basename(prog_name), subcommand),
            description=self.help() or None,
            **kwargs
        )
        parser.add_argument('--version', action='version', version=self.get_version())
        parser.add_argument(
            '--settings',
            help=(
                'project path to a settings module, e.g. '
                '"myproject.settings.py". If this isn\'t provided, the '
                f'{ENV_VAR} environment variable will be used.'
            ),
        )
        self.add_arguments(parser)
        return parser

    def add_arguments(self, parser):
        """
        add custom arguments to the parser.
        """
        pass

    def print_help(self, prog_name, subcommand):
        """
        Print the help message for this command.
        """
        parser = self.create_parser(prog_name, subcommand)
        parser.print_help()

    def run_from_argv(self, argv):
        """
        Set up any environment changes requested.
        """
        parser = self.create_parser(argv[0], argv[1])
        options = parser.parse_args(argv[2:])
        
        cmdline_options = vars(options)
        # Move positional args out of options to mimic legacy optparse
        args = cmdline_options.pop('args', ())
        handle_default_options(options)
        
        self.execute(*args, **cmdline_options)
    
    def handle(self, *args, **options):
        """
        The actual logic of the command. Subclasses must implement this method.
        """
        raise NotImplementedError('subclasses of BaseCommand must provide a handle() method')
    
    def execute(self, *args, **options):
        output = self.handle(*args, **options)
        if output:
            sys.stdout.write(output)
        return output



# import price_val_engine
from os.path import join, dirname

def get_version():
   try:
      with open(join(dirname(__file__), 'price_val_engine/VERSION'), 'rb') as f:
         version = f.read().decode('ascii').strip()
         return version
   except:
      pass


from importlib import import_module

def is_number(value):
    try:
        float(value)
        return True
    except:
        return False
    
    
def import_model(name:str):
    components = name.split('.')
    Klass = components.pop()
    module = import_module(".".join(components))
    return getattr(module, Klass)


import os
# inlude validation pipeline here

# EMPTY VALUES configuration

EMPTY_VALUES = (None, "", [], (), {})

# General Validation Out of range configuration

GEN_VAL_OUT_OF_RANGE_VALUE = {
  'min_value': 100000.0,
  'max_value': 10000000.0
}

# Validation Pipelies
VALIDATION_PIPELINES = [
  'price_val_engine.core.validations.general_rules.NullNegativeZeroValidationRule',
  'price_val_engine.core.validations.general_rules.OutOfRangeValidationRule',
  'price_val_engine.core.validations.revision_rules.DeltaPercentageFromLastDayRule',
]

# AWS CONFIGRATION
AWS_ACCESS_KEY_ID     = os.environ.get('AWS_ACCESS_KEY_ID')
AWS_SECRET_ACCESS_KEY = os.environ.get('AWS_SECRET_ACCESS_KEY')
AWS_SESSION_TOKEN     = os.environ.get('AWS_SESSION_TOKEN')

# google cloud CONFIGRATION
GCS_TOKEN = os.environ.get('GCS_TOKEN', 'anon')


# NOTIFICATION ALEERT
SLACK_ENABLED = False
SLACK_BOT_TOKEN = os.environ.get("SLACK_BOT_TOKEN")
SLACK_CHANNEL = os.environ.get("SLACK_CHANNEL")
# default all
COLUMNS_TO_BE_SEND_ON_ALERT = '__all__'


import os

if __name__ == "__main__":
    from price_val_engine import cmd
    cmd.run_from_commandline()
    
    
    # engine = engine.Engine(
    #     input_file_path="input.csv",
    #     output_file_path="output.csv"
    # )
    # engine.validate_all()
    # engine.save()




import os, stat
import string
from importlib import import_module
from shutil import ignore_patterns, copy2, copystat, move
from argparse import ArgumentError

from price_val_engine.commands import BaseCommand


class Command(BaseCommand):
    
    def add_arguments(self, parser):
        parser.add_argument('name', help='Name of the application or project.')
        parser.add_argument('directory', nargs='?', help='Optional destination directory')
    
    def validate_projectname(self, projectname):
        if projectname is None:
            raise ArgumentError('you must provide project "name"'.format(name=projectname))
        # Check it's a valid project directory name.
        if not projectname.isidentifier():
            raise ArgumentError("'{name}' is not a valid project name." 
                                "Please make sure the name begin with letter " 
                                "and may contains number[0-9] and underscore only".format(name=projectname))
        # Check it cannot be imported.
        try:
            import_module(projectname)
        except ImportError:
            pass
        else:
            raise ArgumentError(
                "'{name}' conflicts with the name of an existing Python module. Please use different name".format(projectname)
            )
    
    @property
    def templates_dir(self):
        import price_val_engine
        return os.path.join(
            os.path.join(price_val_engine.__path__[0], 'conf'),
            'project_template'
        )
    
    def render_templatefile(self, path, **kwargs):
        with open(path, 'rb') as fp:
            raw = fp.read().decode('utf8')
        content = string.Template(raw).substitute(**kwargs)
        render_path = path[:-len('.tmpl')] if path.endswith('.tmpl') else path
        
        if path.endswith('.tmpl'):
            os.rename(path, render_path)
        
        with open(render_path, 'wb') as fp:
            fp.write(content.encode('utf8'))
            
        
    def set_write_permission(self, path):
        os.chmod(path, os.stat(path).st_mode | stat.S_IWUSR)
    
    def copy_templates(self, template_dir, project_dir, project_name):
        file_names = os.listdir(template_dir)
        ignore_files_func = ignore_patterns('*.pyc', '__pycache__', '.svn')
        ignore_files = ignore_files_func(template_dir, file_names)
        
        if not os.path.exists(project_dir):
            os.makedirs(project_dir)

        for file_name in file_names:
            if file_name not in ignore_files:            
                src = os.path.join(template_dir, file_name)
                #print("src =>", src)
                dst = os.path.join(project_dir, file_name)
                # print("dst =>", dst)
                if os.path.isdir(src):
                    self.copy_templates(src, dst, project_name=project_name)
                else:
                    copy2(src, dst)
                    self.set_write_permission(dst)
                    # rename files
                    self.render_templatefile(dst, project_name=project_name)
                    #if dst.endswith('.tmpl'):
                    #    new_file_path =  dst[:-len('.tmpl')]
                    #    os.rename(dst, new_file_path)
                    
        # copy PERNISSIONS
        copystat(template_dir, project_dir)
        self.set_write_permission(project_dir)
        
    def handle(self, *args, **options):
        project_name = options.pop('name')
        target = options.pop('directory')
        self.validate_projectname(project_name)
        if target is None:
            project_dir = os.path.join(os.getcwd(), str(project_name).replace("_", "-"))        
        else:
            project_dir = os.path.abspath(os.path.expanduser(target))
            if not os.path.exists(project_dir):
                raise ArgumentError("Destination directory '%s' does not exist." % project_dir)
        self.copy_templates(self.templates_dir, project_dir, project_name=project_name)
        # rename project_name foler with actual project_name
        move(os.path.join(project_dir, 'project_name'), os.path.join(project_dir, project_name))
        print("New [price_val_engine] Project created.")
    
    


import csv
from io import StringIO
import fsspec
from price_val_engine.conf import settings
from price_val_engine.core.data import FSObject
import boto3


class BaseWriter(FSObject):
    
    def write(self):
        raise NotImplemented("")


class CSVWriter(BaseWriter):
    
    def __init__(self, file_path, mode='wt', storage_options=None, *args, **kwargs) -> None:
        super().__init__(file_path)
        self.args = args
        self.kwargs = kwargs
        self.mode = mode
        self.storage_options = storage_options or {}
        
    def write(self, items, headers):
        filesystem = self.get_fs()
        with filesystem.open(self.file_path, self.mode) as fs:
            writer = csv.DictWriter(fs, fieldnames=headers, *self.args, **self.kwargs)
            writer.writeheader()
            writer.writerows(items)
            fs.close()


import os
import fsspec
from price_val_engine.conf import settings

class FSObject:
    
    def __init__(self, file_path) -> None:
        print(file_path)
        self.file_path = file_path
    
    def get_storage_options(self, protocol):
        storage_options = {}
        if protocol == 's3':
            key = settings.AWS_ACCESS_KEY_ID or os.environ.get('AWS_ACCESS_KEY_ID')
            secret = settings.AWS_SECRET_ACCESS_KEY or os.environ.get('AWS_SECRET_ACCESS_KEY')
            token = settings.AWS_SESSION_TOKEN or os.environ.get('AWS_SESSION_TOKEN')
            storage_options = { 'key': key, 'secret': secret }     
            if token:
                storage_options['token'] = token
        elif protocol == 'gcs':
            token = settings.GCS_TOKEN or os.environ.get('GCS_TOKEN') or 'anon'
            storage_options['token'] = token 
        return storage_options
    
    def get_fs(self) -> fsspec.filesystem:
        protocol, path = fsspec.core.split_protocol(self.file_path)
        storage_options = self.get_storage_options(protocol=protocol)
        return fsspec.filesystem(protocol, **storage_options)
    


import datetime
import socket
import warnings
from price_val_engine.conf import settings
from price_val_engine.core.data import reader, writer
from price_val_engine.core.exceptions import ImproperlyConfigured
from price_val_engine.core import utils

try:
    from slack_sdk import WebClient
except Exception as e:
    print(e)
    warnings.warn("Missing library slack-sdk. alert would not be send")
    WebClient = None
    
class BaseValidationEngine(object):
    validation_rules = settings.VALIDATION_PIPELINES
    data_reader = None
    data_writer = None
    
    def __init__(self, 
        input_file_path, 
        output_file_path,
        storage_options=None,
        validation_rules=[]
        ) -> None:
        
        self.input_file_path = input_file_path
        self.output_file_path = output_file_path
        self.storage_options = storage_options or {}
        # overide validation rules 
        if isinstance(validation_rules, (list, tuple)) and len(validation_rules):
            self.validation_rules = validation_rules
        self.items = []
        
        self.slack_client = None
        if settings.SLACK_ENABLED and WebClient:
            self.slack_client = WebClient(settings.SLACK_BOT_TOKEN)
    
    
    def __slack_message(self, item, error):
        return [
        {
	        "mrkdwn_in": ["text"],
            "color": "#d12815",
            "author_name": f"{socket.getfqdn()}({socket.gethostbyname(socket.gethostname())})",
            "title": f"{error.get('category')}",
            "fields": [
                {
                    "title": "Severity",
                    "value": f"*`{error.get('severity')}`*",
                    "short": True
                },
                {
                    "title": "Message",
                    "value": f"`{error.get('reason')}`",
                    "short": False
                },
                { 
                    "title": "Data",
                    "value": f"{str(self.get_data_on_alert(item))}",
                    "short": False
                }
            ],
            "thumb_url": "https://www.cars24.com/js/28776e9c38260ac3339c3babe6171dd0.svg",
            "footer": "LP Revision",
            "footer_icon": "https://platform.slack-edge.com/img/default_application_icon.png",
            "ts": datetime.datetime.now().timestamp()
        }
    ]
    def get_data_on_alert(self, item:dict):
        if not hasattr(settings, 'COLUMNS_TO_BE_SEND_ON_ALERT'):
            return item
        if settings.COLUMNS_TO_BE_SEND_ON_ALERT in [None, '', '__all__', [], (), {}]:
            return item
        if isinstance(settings.COLUMNS_TO_BE_SEND_ON_ALERT, (list, tuple)):
                return item
        return {k:v for k, v in item.items() if k in settings.COLUMNS_TO_BE_SEND_ON_ALERT}
        
        
    def alert(self, item, error, slacK_channel=None):
        if self.slack_client and error.get('severity') == 'HIGH':
            slacK_channel = slacK_channel or settings.SLACK_CHANNEL
            if slacK_channel == None:
                raise ImproperlyConfigured(
                    "SLACK_CHANNEL should not be blank if SLACK_ENABLED in settings file"
                )
            self.slack_client.chat_postMessage(
                channel=slacK_channel,
                text="Price Validation Engine - alert",
                attachments=self.__slack_message(item, error)
            )
        
    def all(self):
        if self.data_reader is None:
            raise ImproperlyConfigured(
                "Invalid Data Reader class !"
            )
        rows = []
        
        for row in self.data_reader(
                    file_path=self.input_file_path
                ).read():
            rows.append(row)
        return rows
        
    def validate(self, row):
        result = [] 
        for validation_cls in self.validation_rules:
            if isinstance(validation_cls, str):
                klass = utils.import_model(validation_cls)
                validation = klass()
            else:
                validation = validation_cls()
                
            if not validation.is_valid(row):
                result.append(validation.errors)
        if len(result):
            return False, result
        return True, {"category": "success", "severity": "SUCCESS",  "reason": "success"}
    
    def validate_all(self):
        for item in self.all():
            is_valid, response = self.validate(item)
            if is_valid:
                self.items.append({**item, 'is_valid': is_valid, **response})
            else:
                for error in response:
                    # hook to send slack alert on High severaity cases
                    try:
                        self.alert(item, error) 
                    except Exception as e: 
                        print(e)
                    self.items.append({**item, 'is_valid': is_valid, **error})
        return self.items
    
    def save(self):
        if self.data_writer is None:
            raise ImproperlyConfigured(
                "Invalid Data Writer class"
            )  
        if len(self.items):
            fieldnames = list(self.items[0].keys())
            self.data_writer(
                file_path=self.output_file_path
            ).write(self.items, headers=fieldnames)

class DataFrameValidationEngine(BaseValidationEngine):
    
    def __init__(self, DataFrame):
        self.DataFrame = DataFrame
        self.DataFrame.columns= self.DataFrame.columns.str.lower()
        self.items = []
        super().__init__(None, None)

    def all(self):
        yield from self.DataFrame.to_dict(orient='records')
    
    def to_dataframe(self):
        return self.DataFrame.__class__(self.items)
    
    def save(self):
        return self.to_dataframe()
    
    
class ValidationEngine(BaseValidationEngine):
    data_reader = reader.CSVReader
    data_writer = writer.CSVWriter
    
    
Engine = ValidationEngine


from argparse import ArgumentError, ArgumentParser
from importlib import import_module
import pkgutil, os, sys
from price_val_engine.commands import (
    BaseCommand, handle_default_options
)


class CommandLineUtility:

    
    def __init__(self, argv) -> None:
        self.argv = argv or sys.argv[:]
        self.prog_name = os.path.basename(self.argv[0])
        
        if self.prog_name == "__main__.py":
            self.prog_name = 'python -m price_val_engine'
            
        
    @property
    def commands(self):
        import price_val_engine
        command_dir = os.path.join(price_val_engine.__path__[0], "commands")
        return {
            name: "price_val_engine.commands" for _, name, is_pkg in pkgutil.iter_modules([command_dir])
                if not is_pkg and not str(name).startswith("_")
        }
    
    def fetch_command(self, name):
        
        commands = self.commands
        
        try:
            module = commands[name]
        except KeyError:
            if not os.environ.get('PRICE_VAL_ENG_SETTINGS_MODULE'):
                sys.stderr.write("No [price_val_engine] settings specified.\n")
                
            sys.stderr.write('Unknown command: %r' % name)
            
            sys.stderr.write("\nType '%s help' for usage.\n" % self.prog_name)
            sys.exit(1)
            
        if isinstance(module, BaseCommand):
            # If the command is already loaded, use it directly.
            return module
        else:
            module = import_module("{}.{}".format(module, name))
            return module.Command()
        
            # sys.stderr.write('Invalid command: %r' % commnd)
        
        
    def run(self):
        try:
            sub_command = self.argv[1]
        except IndexError:
            sub_command = 'help'  # Display help if no arguments were given.
        
        parser = ArgumentParser(usage='%(prog)s sub_command [options] [args]', add_help=False, allow_abbrev=False)
        parser.add_argument('--settings')
        parser.add_argument('args', nargs='*')  # catch-all
        try:
            options, args = parser.parse_known_args(self.argv[2:])
            handle_default_options(options)
        except ArgumentError:
            pass  # Ignore any option errors at this point.
        
        # _parser = self.fetch_command('run').create_parser('price_val_engine', 'runserver')
        # _options, _args = _parser.parse_known_args(self.argv[2:])
        #for _arg in _args:
        #    self.argv.remove(_arg)


        if sub_command == 'help':
            if '--commands' in args:
                sys.stdout.write("\n".join(sorted(self.commands))  + '\n')
            elif not options.args:
                sys.stdout.write("\n".join(sorted(self.commands))  + '\n')
            else:
                self.fetch_command(options.args[0]).print_help(self.prog_name, options.args[0])
        elif sub_command == 'version' or self.argv[1:] == ['--version'] or self.argv[1:] == ['-v']:
            import price_val_engine
            price_val_engine.get_version()
        else:
            self.fetch_command(sub_command).run_from_argv(self.argv)


def run_from_commandline(argv=None):
    utility = CommandLineUtility(argv)
    utility.run()


import importlib
import os
import warnings
from price_val_engine.conf import default_settings
from price_val_engine.core.exceptions import ImproperlyConfigured

__all__ = [
    "settings",   
]

ENV_VAR = 'PRICE_VAL_ENG_SETTINGS_MODULE'

class Settings:
    def __init__(self, settings_module):
        # update this dict from global settings (but only for ALL_CAPS settings)
        for setting in dir(default_settings):
            if setting.isupper():
                setattr(self, setting, getattr(default_settings, setting))

        # store the settings module in case someone later cares
        self.SETTINGS_MODULE = settings_module
        self._explicit_settings = set()
        
        if self.SETTINGS_MODULE:
            mod = importlib.import_module(self.SETTINGS_MODULE)
            tuple_settings = (
                "VALIDATION_PIPELINES",
            )
            proj_settings = [setting for setting in dir(mod) if not str(setting).startswith("_")]
            for setting in proj_settings:
                if setting.isupper():
                    setting_value = getattr(mod, setting)

                    if setting in tuple_settings and not isinstance(
                        setting_value, (list, tuple)
                    ):
                        raise ImproperlyConfigured(
                            "The %s setting must be a list or a tuple." % setting
                        )
                    setattr(self, setting, setting_value)
                    self._explicit_settings.add(setting)
                    
    def __repr__(self):
        return '<%(cls)s "%(settings_module)s">' % {
            "cls": self.__class__.__name__,
            "settings_module": self.SETTINGS_MODULE,
        }

settings = Settings(os.environ.get(ENV_VAR))




import csv
from price_val_engine.core.data import FSObject

class BaseReader(FSObject):
    
    def read(self):
        raise NotImplemented("")


class CSVReader(BaseReader):
    
    def __init__(self, file_path, mode='rt', *args, **kwargs) -> None:
        super().__init__(file_path)
        self.args = args
        self.kwargs = kwargs
        self.mode = mode
            
    def read(self):
        filesystem = self.get_fs()
        with filesystem.open(self.file_path, self.mode) as fs:
            for row in csv.DictReader(fs, *self.args, **self.kwargs):
                yield row
            fs.close()
                    


from setuptools import find_packages, setup
from os.path import join, dirname

with open(join(dirname(__file__), 'price_val_engine/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()

with open('README.md') as f:
    README = f.read()
    
setup(
    name='price-val-engine',
    packages=find_packages(),
    version=version,
    description='last price revision validation rules',
    long_description=README,
    long_description_content_type= 'text/markdown',
    author='Chandan Kumar Ojha',
    author_email="mr.chandanojha@gmail.com",
    include_package_data=True,
    license='MIT',
    python_requires=">=3.6",
    install_requires=[
        'boto3>=1.21.45',
        'botocore>=1.24.45',
        'fsspec>=2022.1.0',
        'slack_sdk>=3.16.2'
    ]
)


import os
import sys
import fsspec

from price_val_engine.commands import BaseCommand


class Command(BaseCommand):
    
    def add_arguments(self, parser):
        parser.add_argument('--input', help='input file path')
        parser.add_argument('--output', nargs='?', help='output file path', default="output.csv")
    
    def _get_protocol(file_path):
        return fsspec.core.strip_protocol()
                
    def handle(self, *args, **options):
        input_file = options.pop("input")
        output_file = options.pop("output")
        
        if not str(input_file).endswith('.csv'):
            sys.stderr.write("Invalid input [file type]. Please enter csv file path \n")    
        if not str(output_file).endswith('.csv'):
            sys.stderr.write("Invalid output file type, Please enter csv file path \n")
            
        if input_file:
            from price_val_engine.core.engine import Engine
            engine = Engine(
                input_file_path=input_file,
                output_file_path=output_file
            )
            engine.validate_all()
            engine.save()
            
            
        
    
    